package engine

import (
	"strings"
	"testing"

	"github.com/stretchr/testify/assert"
	"golang.org/x/net/html"
)

func TestNewEngine(t *testing.T) {
	e := New(nil)
	assert.NotNil(t, e)
}

func TestProcessNode(t *testing.T) {

	testData := []struct {
		html      string
		hasErrors bool
		expected  string
		data      map[any]any
	}{
		{
			html:      `<div data-test="test"></div>`,
			hasErrors: false,
			expected:  `<div data-test="test"></div>`,
			data: map[any]any{
				"test": "test",
			},
		},
		{
			html:      `<div data-test="test"></div>`,
			hasErrors: false,
			expected:  `<div data-test="test"></div>`,
			data: map[any]any{
				"test": "test",
			},
		},
	}

	for _, td := range testData {
		e := New(td.data)
		doc, _ := html.Parse(strings.NewReader(td.html))
		e.ProcessNode(doc)
		assert.Equal(t, td.hasErrors, e.HasErrors())
		//assert.Equal(t, td.expected, html.Render(doc))
	}

}

func TestProcessHtml(t *testing.T) {

	testData := []struct {
		html      string
		hasErrors bool
		expected  string
		data      map[any]any
	}{
		{
			html:      `<p><span id="1425" data-replace-self="path:text">dummy</span></p>`,
			hasErrors: false,
			expected:  `<p>new text</p>`,
			data: map[any]any{
				"text": "new text",
			},
		},
		{
			html:      `<!DOCTYPE html><html lang="en" data-attributes="lang path:lang"><head><body>`,
			hasErrors: false,
			expected:  `<!DOCTYPE html><html lang="en"><head></head><body></body></html>`,
			data: map[any]any{
				"lang": "en",
			},
		},
		{
			html:      `<div data-test="test"></div>`,
			hasErrors: false,
			expected:  `<div data-test="test"></div>`,
			data:      map[any]any{},
		},
		{
			html:      `<div data-test="test"></div>`,
			hasErrors: false,
			expected:  `<div data-test="test"></div>`,
			data:      map[any]any{},
		},
		{
			html:      `<li>test1</li><li>test2</li>`,
			hasErrors: false,
			expected:  `<li>test1</li><li>test2</li>`,
			data:      map[any]any{},
		},
		{
			html:      `<tr><td>test1</td><td>test2</td></tr><tr><td>test1</td><td>test2</td></tr>`,
			hasErrors: false,
			expected:  `<tr><td>test1</td><td>test2</td></tr><tr><td>test1</td><td>test2</td></tr>`,
			data:      map[any]any{},
		},
		{
			html:      `<td>test1</td><td>test2</td>`,
			hasErrors: false,
			expected:  `<td>test1</td><td>test2</td>`,
			data:      map[any]any{},
		},
	}

	for _, td := range testData {
		e := New(td.data)

		stringWriter := &strings.Builder{}
		e.ProcessHtml(stringWriter, strings.NewReader(td.html))
		processedDoc := stringWriter.String()
		assert.Equal(t, td.hasErrors, e.HasErrors())
		assert.Equal(t, td.expected, processedDoc)

	}

}

func BenchmarkParse(b *testing.B) {
	for i := 0; i < b.N; i++ {
		e := New(nil)
		doc, _ := html.Parse(strings.NewReader(`<div data-test="test"></div>`))
		e.ProcessNode(doc)
	}
}

func TestSetLogNode(t *testing.T) {
	e := New(nil)
	doc, _ := html.Parse(strings.NewReader(`<div data-test="test"></div>`))
	e.SetLogNode(doc.FirstChild)
	assert.True(t, doc.FirstChild.Type == html.ElementNode)
}

func TestSetLogNodeNil(t *testing.T) {
	e := New(nil)
	e.SetLogNode(nil)
}

func TestSetAttributePrefix(t *testing.T) {
	e := New(nil)
	e.SetAttributePrefix("test")
	assert.Equal(t, "test", e.attributePrefix)
}

func TestGetAttributePrefix(t *testing.T) {
	e := New(nil)
	e.SetAttributePrefix("test")
	assert.Equal(t, "test", e.GetAttributePrefix())
}

func TestProcessNodesSingleForDebug(t *testing.T) {

	html := `<div><div data-repeat="city path:cities"><p data-replace="path:city.name"></p></div></div>`
	expected := `<div><div><p>London</p></div><div><p>Paris</p></div></div>`

	mapData := map[any]any{
		"cities": []map[any]any{
			map[any]any{
				"name":       "London",
				"population": 1000000,
			},
			map[any]any{
				"name":       "Paris",
				"population": 2000000,
			},
		},
	}

	e := New(mapData)

	stringWriter := &strings.Builder{}
	e.ProcessHtml(stringWriter, strings.NewReader(html))
	processedDoc := stringWriter.String()
	assert.Equal(t, expected, processedDoc)

	if e.HasErrors() {
		for _, err := range e.GetErrors() {
			t.Log(err)
		}
		t.Fail()
	}

}

func TestProcessNodes(t *testing.T) {

	testData := []struct {
		html      string
		hasErrors bool
		expected  string
	}{
		{
			html:      `<div><div data-repeat="city path:cities"><p data-replace="path:city.name"></p></div></div>`,
			hasErrors: false,
			expected:  `<div><div><p>London</p></div><div><p>Paris</p></div></div>`,
		},
		{
			html:      `<div><div data-replace="path:test"></div></div>`,
			hasErrors: false,
			expected:  `<div><div>test</div></div>`,
		},
		{
			html:      `<div><div data-replace-self="path:test"></div></div>`,
			hasErrors: false,
			expected:  `<div>test</div>`,
		},
		{
			html:      `<div><div data-attributes="class path:test | toupper"></div></div>`,
			hasErrors: false,
			expected:  `<div><div class="TEST"></div></div>`,
		},
		{
			html:      `<div><div data-attributes="class path:test | toupper, style path:cities.0.name | lowercase"></div></div>`,
			hasErrors: false,
			expected:  `<div><div class="TEST" style="london"></div></div>`,
		},
		{
			html:      `<div><div data-repeat="sub path:list"><p data-replace-self="index:sub"></p></div></div>`,
			hasErrors: false,
			expected:  `<div><div>test1</div><div>test2</div><div>test3</div></div>`,
		},
	}

	for _, td := range testData {

		t.Run("test", func(t *testing.T) {

			mapData := map[any]any{
				"test": "test",
				"cities": []map[any]any{
					map[any]any{
						"name":       "London",
						"population": 1000000,
					},
					map[any]any{
						"name":       "Paris",
						"population": 2000000,
					},
				},
				"list": []string{"test1", "test2", "test3"},
			}

			e := New(mapData)

			stringWriter := &strings.Builder{}
			e.ProcessHtml(stringWriter, strings.NewReader(td.html))
			processedDoc := stringWriter.String()
			assert.Equal(t, td.expected, processedDoc)

			if e.HasErrors() {
				for _, err := range e.GetErrors() {
					t.Log(err)
				}
				t.Fail()
			}

		})
	}

}

func BenchmarkProcessNodes(b *testing.B) {

	mapData := map[any]any{
		"test": "test",
		"cities": []map[any]any{
			map[any]any{
				"name":       "London",
				"population": 1000000,
			},
			map[any]any{
				"name":       "Paris",
				"population": 2000000,
			},
		},
	}

	for i := 0; i < b.N; i++ {
		e := New(mapData)
		stringWriter := &strings.Builder{}
		e.ProcessHtml(stringWriter, strings.NewReader(`<div><div data-repeat="city path:cities"><p data-replace="path:city.name"></p></div></div>`))
	}

}

func TestProcessNodesNil(t *testing.T) {
	e := New(nil)
	e.ProcessNodes(nil)
}
